/* SPDX-License-Identifier: GPL-3.0-or-later */
/* Copyright (C) 2021 KikooDX */

#include "conf.h"
#include "game_state.h"
#include "level.h"
#include "particles.h"
#include "player.h"
#include "titlescreen.h"
#include "transition.h"
#include <gint/clock.h>
#include <gint/display.h>
#include <gint/gint.h>
#include <gint/keyboard.h>
#include <gint/timer.h>

#define PANIC(msg)                                                     \
	do {                                                           \
		dclear(C_BLACK);                                       \
		dprint_opt(DWIDTH / 2, DHEIGHT / 2, C_RED, C_NONE,     \
		           DTEXT_CENTER, DTEXT_MIDDLE, "ERROR: %s",    \
		           msg);                                       \
		dupdate();                                             \
		getkey();                                              \
		return 0;                                              \
	} while (0);
#define LOAD_LEVEL()                                                   \
	do {                                                           \
		gint_switch(level_load);                               \
		if (fatal_error == -1)                                 \
			PANIC(fatal_error_msg);                        \
		particles_init();                                      \
		player = player_init();                                \
	} while (0);

extern struct Level level;
extern int level_id;
extern int fatal_error;
extern char *fatal_error_msg;

extern const font_t font_main;

static int callback(volatile void *arg);

int main(void)
{
	int i;
	int timer;
	int player_return_code;
	int level_pack_beaten;
	enum TransitionMode transition_previous_mode;
	enum GameState game_state = TitleScreen;
	volatile int has_ticked = 1;

	struct TitleScreen titlescreen = titlescreen_init();

	/* set font */
	dfont(&font_main);

	/* load level */
	level_id = 0;
	gint_switch(level_load);
	if (fatal_error == -1)
		PANIC(fatal_error_msg);
	/* create player */
	struct Player player = player_init();
	/* initialize particle engine */
	particles_init();
	/* initialize transitions */
	struct Transition transition =
	    transition_init(0.0, C_BLACK, TransitionNone);
	/* timer setup */
	timer = timer_setup(TIMER_ANY, 1000000 / TARGET_UPS, callback,
	                    &has_ticked);
	if (timer == -1)
		PANIC("timer_setup failed");
	timer_start(timer);

	/* main game loop */
	while (!keydown(KEY_EXIT)) {
		/* do x updates per frame, x being ups/fps */
		i = TARGET_UPS / TARGET_FPS;
		while (i-- > 0) {
			/* speed limiter */
			while (!has_ticked)
				sleep();
			has_ticked = 0;

			/* update */
			clearevents();
			switch (game_state) {
			case TitleScreen:
				if (titlescreen_update(&titlescreen))
					game_state = MainMenu;
				break;
			case MainMenu:
				game_state = LevelSelection;
				break;
			case LevelSelection:
				game_state = Playing;
				break;
			case Playing:
				if (transition.mode == TransitionNone) {
					particles_update();
					player_return_code =
					    player_update(&player);
					switch (player_return_code) {
					case 1:
						level_id += 1;
						transition =
						    transition_init(
							H_TRANS_SPD,
							C_BLUE,
							TransitionHOut);
						break;
					case -1:
						transition =
						    transition_init(
							V_TRANS_SPD,
							C_RED,
							TransitionVOut);
						break;
					default:
						break;
					}
				}

				transition_previous_mode =
				    transition_update(&transition);
				if (transition_previous_mode !=
				        transition.mode &&
				    transition_previous_mode !=
				        TransitionNone) {
					switch (
					    transition_previous_mode) {
					case TransitionNone:
						break;
					case TransitionHIn:
						break;
					case TransitionHOut:
						/* end level pack */
						if (level_id % 4 == 0) {
							game_state =
							    PackDone;
							level_pack_beaten =
							    level_id / 4;
						}
						LOAD_LEVEL();
						transition =
						    transition_init(
							transition
							    .speed,
							transition
							    .color,
							TransitionHIn);
						break;
					case TransitionVIn:
						break;
					case TransitionVOut:
						LOAD_LEVEL();
						transition =
						    transition_init(
							transition
							    .speed,
							transition
							    .color,
							TransitionVIn);
						break;
					default:
						break;
					}
				}
				break;
			case PackDone:
				game_state = LevelSelection;
				break;
			default:
				PANIC(
				    "missing game_state case (update)");
				break;
			}

			/* draw */
			dclear(C_BLACK);
			switch (game_state) {
			case TitleScreen:
				titlescreen_draw(titlescreen);
				break;
			case MainMenu:
				break;
			case LevelSelection:
				break;
			case Playing:
				level_draw();
				particles_draw();
				player_draw(player);
				transition_draw(transition);
				break;
			case PackDone:
				break;
			default:
				PANIC("missing game_state case (draw)");
				break;
			}
			dupdate();
		}
	}

	timer_stop(timer);
	return 1;
}

static int callback(volatile void *arg)
{
	volatile int *has_ticked = arg;
	*has_ticked = 1;
	return 0;
}
